@@ -1,13 +1,19 @@
+{-# language CPP #-}
+
+{-# language BangPatterns #-}
 {-# language ExtendedDefaultRules #-}
 {-# language ScopedTypeVariables #-}
 {-# language TupleSections #-}
+{-# language ViewPatterns #-}
 
 module Main (main) where
 
 import Data.Bifunctor (bimap, first, second)
+import qualified Data.Either as Either
 import Data.Function (on)
 import Data.Functor.Identity
 import qualified Data.List as List
+import qualified Data.Maybe as Maybe
 import Data.Ord (Down(..))
 
 import Test.Tasty
@@ -17,9 +23,21 @@ import qualified Data.PQueue.Max as Max
 import qualified Data.PQueue.Min as Min
 import qualified Data.PQueue.Prio.Max as PMax
 import qualified Data.PQueue.Prio.Min as PMin
+import qualified Validity.PQueue.Min as VMin
+import qualified Validity.PQueue.Prio.Min as VPMin
+import qualified Validity.PQueue.Prio.Max as VPMax
 
 default (Int)
 
+validMinQueue :: Ord a => Min.MinQueue a -> Property
+validMinQueue q = VMin.validShape q .&&. VMin.validSize q .&&. VMin.validOrder q
+
+validPMinQueue :: Ord k => PMin.MinPQueue k a -> Property
+validPMinQueue q = VPMin.validShape q .&&. VPMin.validSize q .&&. VPMin.validOrder q
+
+validPMaxQueue :: Ord k => PMax.MaxPQueue k a -> Property
+validPMaxQueue q = VPMax.validShape q .&&. VPMax.validSize q .&&. VPMax.validOrder q
+
 main :: IO ()
 main = defaultMain $ testGroup "pqueue"
   [ testGroup "Data.PQueue.Min"
@@ -28,11 +46,33 @@ main = defaultMain $ testGroup "pqueue"
       [ testProperty "empty" $ Min.getMin Min.empty === Nothing
       , testProperty "non-empty" $ \(NonEmpty xs) -> Min.getMin (Min.fromList xs) === Just (minimum xs)
       ]
-    , testProperty "minView" $ \xs -> Min.minView (Min.fromList xs) === fmap (second Min.fromList) (List.uncons (List.sort xs))
+    , testProperty "minView" $ \xs -> case Min.minView (Min.fromList xs) of
+        Nothing -> xs === []
+        Just (the_min, xs') ->
+           validMinQueue xs' .&&.
+           the_min : Min.toList xs' === List.sort xs
     , testProperty "insert" $ \x xs -> Min.insert x (Min.fromList xs) === Min.fromList (x : xs)
     , testProperty "union" $ \xs ys -> Min.union (Min.fromList xs) (Min.fromList ys) === Min.fromList (xs ++ ys)
-    , testProperty "filter" $ \xs -> Min.filter even (Min.fromList xs) === Min.fromList (List.filter even xs)
-    , testProperty "partition" $ \xs -> Min.partition even (Min.fromList xs) === bimap Min.fromList Min.fromList (List.partition even xs)
+    , testProperty "filter" $ \xs ->
+        let xs' = Min.filter even (Min.fromList xs)
+        in validMinQueue xs' .&&.
+           Min.toList xs' === List.sort (List.filter even xs)
+    , testProperty "partition" $ \xs ->
+        let xs' = Min.fromList xs
+            (ys, zs) = Min.partition even xs'
+        in validMinQueue ys .&&.
+           validMinQueue zs .&&.
+           (Min.toList ys, Min.toList zs) === bimap List.sort List.sort (List.partition even xs)
+    , testProperty "mapMaybe" $ \(Fn f) xs ->
+        let xs' :: Min.MinQueue Char
+            xs' = Min.mapMaybe f (Min.fromList xs)
+        in validMinQueue xs' .&&.
+           Min.toList xs' === List.sort (Maybe.mapMaybe f xs)
+    , testProperty "mapEither" $ \(Fn f) xs ->
+        let (ys, zs) = Min.mapEither f (Min.fromList xs)
+        in validMinQueue ys .&&.
+           validMinQueue zs .&&.
+           (Min.toList ys, Min.toList zs) === bimap List.sort List.sort (Either.partitionEithers . List.map f $ xs)
     , testProperty "map" $ \xs -> Min.map negate (Min.fromList xs) === Min.fromList (List.map negate xs)
     , testProperty "take" $ \n xs -> Min.take n (Min.fromList xs) === List.take n (List.sort xs)
     , testProperty "drop" $ \n xs -> Min.drop n (Min.fromList xs) === Min.fromList (List.drop n (List.sort xs))
@@ -48,7 +88,14 @@ main = defaultMain $ testGroup "pqueue"
     , testProperty "toDescList" $ \xs -> Min.toDescList (Min.fromList xs) === List.sortOn Down xs
     , testProperty "fromAscList" $ \xs -> Min.fromAscList (List.sort xs) === Min.fromList xs
     , testProperty "fromDescList" $ \xs -> Min.fromDescList (List.sortOn Down xs) === Min.fromList xs
-    , testProperty "mapU" $ \xs -> Min.mapU (+ 1) (Min.fromList xs) === Min.fromList (List.map (+ 1) xs)
+    , testProperty "mapU" $ \xs ->
+        let
+          -- Monotonic, but not strictly so
+          fun x
+            | even x = x
+            | otherwise = x + 1
+          res = Min.mapU fun (Min.fromList xs)
+        in validMinQueue res .&&. Min.toList res === List.map fun (List.sort xs)
     , testProperty "foldrU" $ \xs -> Min.foldrU (+) 0 (Min.fromList xs) === sum xs
     , testProperty "foldlU" $ \xs -> Min.foldlU (+) 0 (Min.fromList xs) === sum xs
     , testProperty "foldlU'" $ \xs -> Min.foldlU' (+) 0 (Min.fromList xs) === sum xs
@@ -106,9 +153,20 @@ main = defaultMain $ testGroup "pqueue"
       [ testProperty "Just" $ \xs -> PMin.updateMinA (Identity . Just) (PMin.fromList xs) === Identity (PMin.fromList xs)
       , testProperty "Nothing" $ \(NonEmpty (xs :: [(Int, ())])) -> PMin.updateMinA (Identity . const Nothing) (PMin.fromList xs) === Identity (PMin.fromList (tail (List.sort xs)))
       ]
-    , testProperty "minViewWithKey" $ \(xs :: [(Int, ())]) -> PMin.minViewWithKey (PMin.fromList xs) === fmap (second PMin.fromList) (List.uncons (List.sort xs))
+    , testProperty "minViewWithKey" $ \(xs :: [(Int, Int)]) -> case PMin.minViewWithKey (PMin.fromList xs) of
+        Nothing -> xs === []
+        Just ((the_min, the_min_val), xs') ->
+           validPMinQueue xs' .&&.
+           List.sort ((the_min, the_min_val) : PMin.toList xs') === List.sort xs
     , testProperty "map" $ \(xs :: [(Int, ())]) -> PMin.map id (PMin.fromList xs) === PMin.fromList xs
-    , testProperty "mapKeysMonotonic" $ \xs -> PMin.mapKeysMonotonic (+ 1) (PMin.fromList xs) === PMin.fromList (List.map (first (+ 1)) xs)
+    , testProperty "mapKeysMonotonic" $ \xs ->
+        let
+          -- Monotonic, but not strictly so
+          fun x
+            | even x = x
+            | otherwise = x + 1
+          res = PMin.mapKeysMonotonic fun (PMin.fromList xs)
+        in validPMinQueue res .&&. List.sort (PMin.toList res) === List.sort (List.map (first fun) xs)
     , testProperty "take" $ \n (xs :: [(Int, ())]) -> PMin.take n (PMin.fromList xs) === List.take n (List.sort xs)
     , testProperty "drop" $ \n (xs :: [(Int, ())]) -> PMin.drop n (PMin.fromList xs) === PMin.fromList (List.drop n (List.sort xs))
     , testProperty "splitAt" $ \n (xs :: [(Int, ())]) -> PMin.splitAt n (PMin.fromList xs) === second PMin.fromList (List.splitAt n (List.sort xs))
@@ -123,10 +181,35 @@ main = defaultMain $ testGroup "pqueue"
       \(Fn2 (f :: Int -> () -> Maybe ())) (xs :: [(Int, ())]) -> PMin.mapMWithKey f (PMin.fromList xs) === fmap PMin.fromList (traverse (\(k, x) -> fmap (k,) (f k x)) xs)
     , testProperty "insert" $ \k xs -> PMin.insert k () (PMin.fromList xs) === PMin.fromList ((k, ()) : xs)
     , testProperty "union" $ \(xs :: [(Int, ())]) ys -> PMin.union (PMin.fromList xs) (PMin.fromList ys) === PMin.fromList (xs ++ ys)
-    , testProperty "filter" $
-      \(xs :: [(Int, ())]) -> PMin.filterWithKey (\k _ -> even k) (PMin.fromList xs) === PMin.fromList (List.filter (even . fst) xs)
-    , testProperty "partition" $
-      \(xs :: [(Int, ())]) -> PMin.partitionWithKey (\k _ -> even k) (PMin.fromList xs) === bimap PMin.fromList PMin.fromList (List.partition (even . fst) xs)
+    , testProperty "filter" $ \(xs :: [(Int, Int)]) ->
+        let
+          -- The probability of a number not being divisible by 3 is 2/3.
+          -- The probability of a number not being divisible by 4 is 3/4.
+          -- So the probability of a number being divisible by neither is
+          -- 1/2.
+          f x y = x `rem` 3 == 0 || y `rem` 4 == 0
+          xs' = PMin.filterWithKey f (PMin.fromList xs)
+        in validPMinQueue xs' .&&.
+           List.sort (PMin.toList xs') === List.sort (List.filter (uncurry f) xs)
+    , testProperty "partition" $ \(xs :: [(Int, Int)]) ->
+        let
+          f x y = x `rem` 3 == 0 || y `rem` 4 == 0
+          (ys, zs) = PMin.partitionWithKey f (PMin.fromList xs)
+        in validPMinQueue ys .&&.
+           validPMinQueue zs .&&.
+           (List.sort (PMin.toList ys), List.sort (PMin.toList zs)) ===
+             bimap List.sort List.sort (List.partition (uncurry f) xs)
+    , testProperty "mapMaybe" $ \(Fn2 f) (xs :: [(Int, Int)]) ->
+        let
+          xs' = PMin.mapMaybeWithKey f (PMin.fromList xs)
+        in validPMinQueue xs' .&&.
+           List.sort (PMin.toList xs') === List.sort (Maybe.mapMaybe (\(k,v) -> fmap (k,) (f k v)) xs)
+    , testProperty "mapEither" $ \(Fn2 f) (xs :: [(Int, Int)]) ->
+        let (ys, zs) = PMin.mapEitherWithKey f (PMin.fromList xs)
+        in validPMinQueue ys .&&.
+           validPMinQueue zs .&&.
+           (List.sort (PMin.toList ys), List.sort (PMin.toList zs)) ===
+             bimap List.sort List.sort (Either.partitionEithers . List.map (\(k,v) -> bimap (k,) (k,) (f k v)) $ xs)
     , testProperty "toAscList" $ \(xs :: [(Int, ())]) -> PMin.toAscList (PMin.fromList xs) === List.sort xs
     , testProperty "toDescList" $ \(xs :: [(Int, ())]) -> PMin.toDescList (PMin.fromList xs) === List.sortOn Down xs
     , testProperty "fromAscList" $ \(xs :: [(Int, ())]) -> PMin.fromAscList (List.sort xs) === PMin.fromList xs
@@ -158,7 +241,14 @@ main = defaultMain $ testGroup "pqueue"
       ]
     , testProperty "minViewWithKey" $ \(xs :: [(Int, ())]) -> PMax.maxViewWithKey (PMax.fromList xs) === fmap (second PMax.fromList) (List.uncons (List.sortOn Down xs))
     , testProperty "map" $ \(xs :: [(Int, ())]) -> PMax.map id (PMax.fromList xs) === PMax.fromList xs
-    , testProperty "mapKeysMonotonic" $ \xs -> PMax.mapKeysMonotonic (+ 1) (PMax.fromList xs) === PMax.fromList (List.map (first (+ 1)) xs)
+    , testProperty "mapKeysMonotonic" $ \xs ->
+        let
+          -- Monotonic, but not strictly so
+          fun x
+            | even x = x
+            | otherwise = x + 1
+          res = PMax.mapKeysMonotonic fun (PMax.fromList xs)
+        in validPMaxQueue res .&&. List.sort (PMax.toList res) === List.sort (List.map (first fun) xs)
     , testProperty "take" $ \n (xs :: [(Int, ())]) -> PMax.take n (PMax.fromList xs) === List.take n (List.sortOn Down xs)
     , testProperty "drop" $ \n (xs :: [(Int, ())]) -> PMax.drop n (PMax.fromList xs) === PMax.fromList (List.drop n (List.sortOn Down xs))
     , testProperty "splitAt" $ \n (xs :: [(Int, ())]) -> PMax.splitAt n (PMax.fromList xs) === second PMax.fromList (List.splitAt n (List.sortOn Down xs))
